/*
**	Barbuceanu Malin-Andrei 333CC
**  Joc Snake
*/
#define LCD_WIDTH  128
#define LCD_HEIGHT 64
#define F_CPU         16000000
#include "assets/fonts.h"
#include "assets/img.h"
#include <avr/io.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <util/delay.h>
#include <avr/eeprom.h>
#include <math.h>

#define INITIAL_SIZE 6
#define MAX_SIZE     50
#define MAX_WALL_HEIGHT 16
#define MAX_WALL_WIDTH 10

#define BLOCK_SIZE 1

#define P_RIGHT PB6
#define P_LEFT PB7
#define P_UP PB5
#define P_DOWN PB4

// directiile
#define D_UP    0
#define D_DOWN  1
#define D_LEFT  2
#define D_RIGHT 3
#define D_NONE  5

// nr de puncte care trebuie acumulat intr un nivel
#define NR_FOOD_TO_PICK 5

// starile jocului
typedef enum {
        WELCOME,
        RUNNING,
        LOST,
        WON,
        READY
} state;

// coordonatele unui punct din mapa
typedef struct {
    uint8_t x;
    uint8_t y;
    uint8_t width;
    uint8_t height;
} coord;

coord snake[MAX_SIZE];  // coordonatele sarpelui

coord food;

coord wall[MAX_SIZE];   // coordontale zidului

uint32_t snake_size;

unsigned int score;
unsigned int highest_score = 0;

// pt generarea random
unsigned int seed = 52;

int time = 90; // timpul in milisecunde intre iteratii

int level = 0;

int ok = 0;

state game_state = WELCOME;
int direction = D_LEFT;

// verificarea apasarii unui buton
unsigned int left_key_pressed() {
        return !(PINB & (1 << P_LEFT));
}

unsigned int right_key_pressed() {
        return !(PINB & (1 << P_RIGHT));
}

unsigned int up_key_pressed() {
        return !(PINB & (1 << P_UP));
}

unsigned int down_key_pressed() {
        return !(PINB & (1 << P_DOWN));
}

// seteaza inputurile si activeaza rezistentele de pull up ale butoanelor
// respectiv ledurilor
void init_input() {
        DDRB &= ~((1 << PB7) | (1 << PB6));
        PORTB |= (1 << PB7) | (1 << PB6);

        DDRB &= ~((1 << PB5) | (1 << PB4));
        PORTB |= (1 << PB5) | (1 << PB4);

        DDRB |= (1 << PB3) | (1 << PB2);
        PORTB &= ~(1 << PB3);
        PORTB &= ~(1 << PB2);

        // initializare difuzor
        DDRD |= (1 << PD5);

        PORTD &= ~(1 << PD5);


        GLI_Initialize();
        GLI_Clear(GLI_COLOR_BLACK);
        GLI_Present();
}

// genereaza coordonatele hranei
void generate_coord(void) {
    
        srand(seed);

        food.x = rand() % (LCD_WIDTH - 3);
        food.y = rand() % (LCD_HEIGHT - 3);

        // vrific sa nu plaseze prada unde nu am acces la ea
        if (food.x < 6)
                food.x += 6;

        if (food.y < 15)
                food.y += 15;

        seed = seed + food.x;
}

// genereaza coordonatele zidului
void generate_walls(void) {

        int i;

        for (i = 0; i < level; i++) {
                srand(seed);

                wall[i].x = rand() % (LCD_WIDTH - 3);
                wall[i].y = rand() % (LCD_HEIGHT - 3);
                wall[i].height = rand() % MAX_WALL_HEIGHT + 1;
                wall[i].width = rand() % MAX_WALL_WIDTH + 1;

                // zidurile nu trebuie generate in zona de start a sarpelui
                if (wall[i].y < 10) 
                        wall[i].y += 10;

				if (wall[i].x < 36 && wall[i].y < LCD_HEIGHT/2 + 1)
					wall[i].x += 36;

                if (wall[i].height <= 2 && wall[i].width <= 2)
                    wall[i].width += 4;
                seed = seed + wall[i].x;
        }
}

void init_game_state() {

        int startx = 3; // pozitia de start
        score = 0;
        game_state = WELCOME;

        direction = D_LEFT;
        snake_size = INITIAL_SIZE;

        int i;

		//initializez pozitiile blocurilor initiale ala sarpelui
        for (i = 0; i < snake_size; i++) {
                snake[i].x = startx + 10 + i * BLOCK_SIZE * 2;
                snake[i].y = LCD_HEIGHT/2;
        }

		// generez coordonatele pt mancare
        generate_coord();
}

// trateaza inputul userilor
void handle_button() {

		// inceputul jocului
        if (right_key_pressed()) {
                if (game_state == WELCOME) {
                        score = 0;
                        GLI_Clear(GLI_COLOR_BLACK);
                        game_state = RUNNING;
                }
                else if (game_state == RUNNING) {
                }
                else {
                        init_game_state();
                        game_state = RUNNING;
                }
        }

		// stabilesc directia de deplasare a sarpelui
        if (right_key_pressed() && (direction == D_UP || direction == D_DOWN))
                direction = D_RIGHT;
        if (left_key_pressed() && (direction == D_UP || direction == D_DOWN))
                direction = D_LEFT;
        if (up_key_pressed() && (direction == D_LEFT || direction == D_RIGHT))
                direction = D_UP;
        if (down_key_pressed() && (direction == D_LEFT || direction == D_RIGHT))
                direction = D_DOWN;

}

// deseneaza un dreptunghi care are coltul stanga sus in (topx,topy), lungimea length
// si latimea width
void draw_line(unsigned int topx, unsigned int topy, unsigned int length, unsigned int width) {
        int i,j;
        for (i=topx; i<=topx+width; i++)
                for (j=topy; j<=topy+length; j++)
                        GLI_SetPixel(i,j,0xff);

}

// preiau coordonatele blocurilor si il reprezint pe fiecare apeland functia draw_line
void draw_snake() {

        int i;
        for (i = 0; i < snake_size; i++) {
                draw_line(snake[i].x, snake[i].y, BLOCK_SIZE, BLOCK_SIZE);
        }
}

// verific daca am gasit mancare
void check_food() {

        int headx = snake[snake_size - 1].x;
        int heady = snake[snake_size - 1].y;
        if (abs(headx - food.x) < 2 && abs(heady - food.y) < 2) {
                generate_coord();
                score++;

                snake_size++;
                int i;

                for (i = snake_size - 1; i > 0; i--) {
                        snake[i].x = snake[i - 1].x;
                        snake[i].y = snake[i - 1].y;
                }

                PORTB |= (1 << PB3);
        }
        else
                PORTB &= ~(1 << PB3);
}

// verific coliziunea sarpelui cu el insusi sau cu obstacole
void check_collision() {

        int headx = snake[snake_size - 1].x;
        int heady = snake[snake_size - 1].y;
        int i;

		// coliziunea cu el insusi
        for (i = 0; i < snake_size - 3; i++) {

                int blockx = snake[i].x;
                int blocky = snake[i].y;
                if (abs(headx - blockx) < 2 && abs(heady - blocky) < 2) {
                        game_state = LOST;

                        PORTB |= (1 << PB2);
                }
                else
                        PORTB &= ~(1 << PB2);
        }

		// coliziunea cu obstacole
        for (i = 0; i < level; i++) {

                if (headx >= wall[i].x && headx <= wall[i].x + wall[i].width) {
                        if (heady >= wall[i].y && heady <= wall[i].y + wall[i].height) {
                                game_state = LOST;

                                PORTB |= (1 << PB2);
                        } else {
                                PORTB &= ~(1 << PB2); 
                        }
                }
        }
}

// actualizez pozitia sarpelui
void compute_coords() {

        int i;
        for (i = 0; i < snake_size - 1; i++){
                snake[i].x = snake[i + 1].x;
                snake[i].y = snake[i + 1].y;
        }

        if (direction == D_RIGHT)
                snake[snake_size - 1].x = snake[snake_size - 1].x - 2 * BLOCK_SIZE;
        else if (direction == D_LEFT)
                snake[snake_size - 1].x = snake[snake_size - 1].x + 2 * BLOCK_SIZE;
        else if (direction == D_DOWN)
                snake[snake_size - 1].y = snake[snake_size - 1].y - 2 * BLOCK_SIZE;
        else    //UP
                snake[snake_size - 1].y = snake[snake_size - 1].y + 2 * BLOCK_SIZE;

        if (direction == D_LEFT) {
                
                for (i = 0; i < snake_size; i++)
                        if (snake[i].x > LCD_WIDTH - 2)
                                snake[i].x = i + 1;
        } else if (direction == D_RIGHT) {
                
                for (i = 0; i < snake_size; i++)
                        if(snake[i].x < 2)
                                snake[i].x = LCD_WIDTH - i - 2;
        } else if (direction == D_UP ) {
                
                for (i = 0; i < snake_size; i++)
                        if(snake[i].y > LCD_HEIGHT)
                                snake[i].y = i + 10;
        } else if (direction == D_DOWN) {
                
                for (i = 0; i < snake_size; i++)
                        if(snake[i].y < 10)
                                snake[i].y = LCD_HEIGHT - i;
        }

		// verific daca s a produs o coliziune
        check_collision();

		// verifica daca am gasit mancare
        check_food();
}

// deseneaza sarpele, mancarea si obstacolele
void draw_map() {

        compute_coords();
        draw_snake();
        draw_line(food.x, food.y, BLOCK_SIZE, BLOCK_SIZE);

        int i;
        for (i = 0; i < level; i++) {
                draw_line(wall[i].x, wall[i].y, wall[i].height, wall[i].width);
        }
}

// afiseaza mesajul de inceput
void draw_welcome_screen() {
        GLI_DrawString(1,20,&lucida_console, "Welcome to Snake");
        GLI_DrawString(1,35,&lucida_console, "Press the button to start");
        GLI_Present();
}

// afisez pe ecranul jocului scorul curent, best score si nivelul
void draw_score() {
        char score_msg[30];
        sprintf(score_msg,"you: %d, best: %d lv: %d", score, highest_score, level);
        GLI_DrawString(0,0,&lucida_console,score_msg);
}

// afisez jocul - harta + scoruri
void draw_game_screen() {
        GLI_Clear(GLI_COLOR_BLACK);
        draw_map();
        draw_score();
        GLI_Present();
}

// determin care este starea jocului si in functie de aceasta desenez ecranul corespunzator
void draw_screen() {
        if (game_state == WELCOME) {
                draw_welcome_screen();
        }
        else if (game_state == RUNNING) {
                ok = 0;
                draw_game_screen();
        }
        else if (game_state == WON) {

                if (ok == 0) {
                        level++;
                        generate_walls();
                        ok = 1;
                        score = 0;

                        if (score + NR_FOOD_TO_PICK * level> highest_score)
                                highest_score = score + NR_FOOD_TO_PICK * level;
                }
                GLI_Clear(GLI_COLOR_BLACK);
                GLI_DrawString(3,30,&lucida_console, "You won!");
                GLI_DrawString(3,40,&lucida_console, "Press the button to try again");
                GLI_Present();

        } else {    // LOST

                if (score + NR_FOOD_TO_PICK * level> highest_score)
                    highest_score = score + NR_FOOD_TO_PICK * level;
                level = 0;
                GLI_Clear(GLI_COLOR_BLACK);
                GLI_DrawString(3,30,&lucida_console, "You lost :(");
                GLI_DrawString(3,40,&lucida_console, "Press the button to try again");
                GLI_Present();
        }
}

void speaker_morse() {

    int durata = 50;
    int i;
    for (i = 0; i < durata; ++i) {
        PORTD ^= (1 << PD5);

        _delay_ms(1);
    }
}

int main() {
        init_input();
        init_game_state();
        generate_coord();
        while (1) {

                handle_button();
                draw_screen();
                
				_delay_ms(time);	// stabileste viteza de deplasare a sarpelui

				// verific daca a acumulat numarul de puncte njecesar trecerii la urmatorul nivel
                if (score == NR_FOOD_TO_PICK)
                        game_state = WON;

              //  speaker_morse();

              //  PORTD &= ~(1 << PD5);
        }
        GLI_Clear(GLI_COLOR_BLACK);
        return 0;
}
